clear all;
clc;

x_min = -3;  % Minimum pro x1 i x2
x_max = +3;  % Maximum pro x1 i x2
x_step = 0.05;  % Krok vzorkování v obou proměnných

%%
% Vzorky pro x1 i x2 jsou v tomto případě totožné
%%% x1 = ...:...:...;
%%% x2 = ...:...:...;

%%
% Matice pro j1(x1,x2) a j2(x1,x2) se vykreslují pomocí pcolor, a mají
% proto tolik _slopupců_. kolik je vzorků na ose x resp. `x1` a tolik
% řádek, kolik je vzorků na ose y resp. `x2`.
%%% z1 = zeros(..., ...);
%%% z2 = zeros(length(x2), length(x1));

% Bylo by výhodnější použít ndgrid() ...
for col = 1:length(x1)
    for row = 1:length(x2)
        %%% x1x2 = [ x1(...), x2(...) ];  % Souřadnice, které počítáme
        %%% z1(...,...) = j1(x1x2);  % První kritérium
        %%% z2(...,...) = j2(x1x2);  % Druhé kritérium
    end
end

% Vykreslíme vedle sebe J1 a J2
figure(1);

subplot(1,2,1);
pcolor(x1,x2,z1);
axis equal;
shading interp;
colormap jet;
xlim([x_min, x_max]);  % Je to čtverec, obě osy jdou od x_min do x_max
ylim([x_min, x_max]);
xlabel('x_1');
ylabel('x_2');
title('J_1(x_1,x_2)');

subplot(1,2,2);
pcolor(x1,x2,z2);
axis equal;
shading interp;
colormap jet;
xlim([x_min, x_max]);
ylim([x_min, x_max]);
xlabel('x_1');
ylabel('x_2');
title('J_2(x_1,x_2)');

% Vykreslíme vedle sebe 0.5*J1 + 0.5*J2
figure(2);
%%% pcolor(x1,x2,...);
axis equal;
shading interp;
colormap jet;
xlim([x_min, x_max]);  % Je to čtverec, obě osy jdou od x_min do x_max
ylim([x_min, x_max]);
xlabel('x_1');
ylabel('x_2');
title('0.5*J_1(x_1,x_2)+0.5*J_2(x_1,x_2)');

% Protože nevíme, kde přesně nastartovat fmincon(), aby našlo globální
% maximum, musíme zkusit opakované starty z určité mřížky bodů
%%% x_starts = linspace(...,...,6);
% Odmažu první a poslední souřadnici, protože ty jsou zcela na hranici
% přípustné oblasti pro x
%%% x_starts = x_starts(...);

% Mřížka startovacích bodů zadaná jako 2D sloupcový vektor
% Kolik těch bodů se souřadnicemi v `x_starts` bude?
% Bylo by výhodnější použít ndgrid() ...
%%% x0_start = zeros(...,2);
i = 1;
for col = 1:length(x_starts)
    for row = 1:length(x_starts)
        % Na i-tý řádek zapíšeme souřadnice, ze kterých v i-tém pokusu
        % nalézt globální optimum startujeme
        %%% x0_start(...) = [ x_starts(...), x_starts(...) ];
        i = i+1;
    end
end

% Vlastní pokus o maximalizaci J1 bez omezení nedopadne dobře ...
%j1_max = -Inf;
%for i = 1:length(x0_start)
%    [j1_maxc] = fmincon(@(x)-j1(x), x0_start(i,:));
%    j1_val = j1(j1_maxc);
%    if j1_val > j1_max
%        j1_max = j1_val;
%        j1_c = j1_maxc;
%    end
%end

% ... a proto omezíme přípustné hodnoty x1 a x2 na interval <x_min,x_max>
LB = [ x_min, x_min ];  % Dolní mez (lower bound)
UB = [ x_max, x_max ];  % Horní mez (upper bound)
% Nastavíme parametry optimalizace: tolerance, vypnutí výpisů na obrazovku
options = optimset('LargeScale', 'off', 'MaxFunEvals', 1000,...
    'TolFun', 1e-6, 'TolCon', 1e-6, 'disp', 'off');
j1_max = -Inf;
% for i = 1:length(x0_start)
%     % Funkce fmincon() má spoustu prametrů, LB a UB jsou až hodně vzadu,
%     % před ně musíme zadat prázdné hodnoty ...
%     [j1_maxc] = fmincon(@(x)-j1(x), x0_start(i,:), [], [], [], [], LB, UB, [], options);
%     j1_val = j1(j1_maxc);
%     if j1_val > j1_max
%         j1_max = j1_val;
%         j1_c = j1_maxc;
%     end
% end

% Vytvoříme vektor vah pro J(x)=w*J1(x)+(1-w)*J2(x), w jde od 0 do 1
num_w = 50;
%%% w = linspace(...,num_w);  % num_w vah rovnoměrně na <0,1>
% Sem si budeme ukládat optima pro jednotlivé váhy. Vektor má num_w řádek a
% dva sloupce pro hodnoty 
%%% xopt_w = zeros(...);
% To samé pro funkce J1 a J2
%%% jopt_w = zeros(...);

% Aproximace pareto-optima pomocí lineární kombinace J1 a J2
for j=1:num_w
    opt_max = -Inf;
    opt_maxc = [];
    % Musíme optimalizovat několikrát, abychom měli šanci, že najdeme
    % globální maximum ...
    for i = 1:length(x0_start)
        % Funkce fmincon() má spoustu prametrů, LB a UB jsou až hodně vzadu,
        % před ně musíme zadat prázdné hodnoty ...
        %%% [opt_c] = fmincon(..., x0_start(...), [], [], [], [], LB, UB, [], options);
        % Máme x1 a x2 bodu, kde vážený součet nabývá maxima, musíme ještě
        % spočíst funkční hodnotu, abychom věděli, zda nemáme větší
        % maximum než to, které jsme doteď našli
        % Pozn.: fmincon umí vrátit funkční hodnotu minima, bude to ale
        % záporná hodnota našeho maxima ... proto možná lépe takto:
        opt_val = w(j)*j1(opt_c) + (1-w(j))*j2(opt_c);
        if opt_val > opt_max
            opt_max = opt_val;
            opt_maxc = opt_c;
        end
    end
    % Nyní už máme snad globální maximum na naší přípustné množině.
    % Uložíme hodnoty.
    xopt_w(j,:) = opt_maxc;
    jopt_w(j,:) = [ j1(opt_maxc), j2(opt_maxc) ];
end

% Vykreslíme prostor párů funkčních hodnot s přibližnou pareto-optimální
% hranicí určenou výše.
% Všechny dvojice funkčních hodnot J1 a J2 máme v maticích z1 a z2, ty lze
% převést na vektory jako z1(:) a z2(:).
figure(3);
plot(z1(:),z2(:),'k+', jopt_w(:,1), jopt_w(:,2), 'rs', jopt_w(:,1), jopt_w(:,2), 'r-');
axis equal;
xlabel('J_1(x_1,x_2)');
ylabel('J_2(x_1,x_2)');
title('Přibližná Pareto-optimální hranice - převod na 1kriteriální úlohu');

% Pokusíme se nalézt nekonvexní Pareto-optimální hranici pomocí vyšetřování
% dominance bodů v z1 a z2
z1flat = z1(:);  % převod matice bodů na vektor/sekvenci
z2flat = z2(:);
idx_dominated = find_dominated(-z1flat,-z2flat);  % hledáme maxima, proto obrat znaménka
% Vybereme ty, které _nejsou_ dominované
z1nd = z1flat(~idx_dominated);
z2nd = z2flat(~idx_dominated);
% Pokud chceme spojnici, musíme vzorky zeřadit například podle hodnot J1
% V Matlabu ovšem sort() řadí _nezávisle_ všechny sloupce matice, musíme
% tedy postupně
[z1s, idx_s] = sort(z1nd);  % seřazeno J1, plus index záměn v J1
z2s = z2nd(idx_s);  % indexem záměn zpřeházíme hodnoty v J2

% Vykreslíme
figure(4);
plot(z1(:),z2(:),'k+', z1s, z2s, 'ms', z1s, z2s, 'm-');
axis equal;
xlabel('J_1(x_1,x_2)');
ylabel('J_2(x_1,x_2)');
title('Přibližná nekonvexní Pareto-optimální hranice');






